import pytest
import xarray as xr
from xrviz.display import Display
from . import data
from ..utils import _is_coord


@pytest.fixture()
def displayer(data):
    return Display(data)


@pytest.fixture()
def displayer_array(data):
    data = data.temp
    data = xr.Dataset({f'{data.name}': data}, attrs=data.attrs) 
    return Display(data)


@pytest.fixture()
def displayer_array_single_var(data):
    data = data.lon
    data = xr.Dataset({f'{data.name}': data}, attrs=data.attrs) 
    return Display(data)


def test_fill_items(displayer):
    assert displayer.select.options == {'validtime': 'validtime',
                                        'time 📈': 'time',
                                        'lon': 'lon',
                                        'lat': 'lat',
                                        'mask': 'mask',
                                        'depth': 'depth',
                                        'sigma 📈': 'sigma',
                                        'zeta': 'zeta',
                                        'air_u': 'air_u',
                                        'air_v': 'air_v',
                                        'u': 'u',
                                        'v': 'v',
                                        'temp': 'temp'}


def test_fill_items_array(displayer_array):
    assert displayer_array.select.options == {'time 📈': 'time',
                                              'sigma 📈': 'sigma',
                                              'temp': 'temp'}


def test_fill_items_single_var(displayer_array_single_var):
    assert displayer_array_single_var.select.options == {'lon': 'lon'}



import pytest
import panel as pn
from ..compatibility import has_cartopy

if not has_cartopy:
    pytest.skip("cartopy geoviews not installed", allow_module_level=True)
else:
    from geoviews import tile_sources as gvts
    from xrviz.projection import Projection, projections_list
    from xrviz.utils import proj_params
    basemap_opts = {None: None}
    basemap_opts.update(gvts.tile_sources)


@pytest.fixture()
def projection():
    return Projection()


@pytest.mark.parametrize("wid, wid_type, attrs_values",
                         [('is_geo', pn.widgets.Checkbox,
                           [('name', 'is_geo'), ('value', False),
                            ('disabled', True)]),
                          ('alpha', pn.widgets.FloatSlider,
                           [('name', 'alpha'), ('start', 0), ('end', 1),
                            ('step', 0.01), ('value', 0.7), ('width', 180)]),
                          ('basemap', pn.widgets.Select,
                           [('name', 'basemap'),
                            ('options', basemap_opts),
                            ('value', None), ('disabled', True)]),
                          ('projection', pn.widgets.Select,
                           [('name', 'projection'),
                            ('options', [None] + projections_list),
                            ('value', None)]),
                          ('crs', pn.widgets.Select,
                           [('name', 'crs'),
                            ('options', sorted(projections_list)),
                            ('value', 'PlateCarree')]),
                          ('project', pn.widgets.Checkbox,
                           [('name', 'project'), ('value', False)]),
                          ('global_extent', pn.widgets.Checkbox,
                           [('name', 'global_extent'), ('value', False)]),
                          ('features', pn.widgets.MultiSelect,
                           [('name', 'features'),
                            ('options', ['None', 'borders', 'coastline',
                                         'grid', 'land', 'lakes',
                                         'ocean', 'rivers']),
                            ('value', ['borders', 'coastline',
                                       'grid', 'land', 'lakes',
                                       'ocean', 'rivers'])]) ])
def test_widget(projection, wid, wid_type, attrs_values):
    assert isinstance(projection.is_geo, pn.widgets.Checkbox)
    widget = getattr(projection, wid)
    assert isinstance(widget, wid_type)
    for attr_val in attrs_values:
        assert getattr(widget, attr_val[0]) == attr_val[1]


def test_setup_geo_disabled(projection):
    disabled_val = True
    projection.is_geo.disabled = True
    projection._emit('geo_disabled', '')
    for row in projection.panel[1:]:
        for widget in row:
            assert widget.disabled is True
    assert projection.basemap.disabled is True


def test_setup_geo_changed_to_False(projection):
    projection.is_geo.disabled = False
    projection.is_geo.value = False
    for row in projection.panel[1:]:
        for widget in row:
            assert widget.disabled is True
    assert projection.basemap.disabled is True


def test_setup_geo_changed_to_True(projection):
    projection.is_geo.disabled = False
    projection.is_geo.value = True
    for row in projection.panel[1:]:
        for widget in row:
            assert widget.disabled is False


def test_add_proj_params(projection):
    projection.is_geo.disabled = False
    sample_projs = ['PlateCarree', 'Orthographic', 'EuroPP']
    for proj_value in sample_projs:
        projection.projection.value = proj_value
        assert projection.proj_params.value == proj_params(proj_value)

        # accepted_params = accepted_proj_params(proj_value)
        # for widget in projection.proj_params:
        #     assert widget.name in accepted_params



import inspect
import panel as pn
from .compatibility import ccrs, has_cartopy


def convert_widget(source, target):
    """
    To convert a widget form one type to another having
    same name, value and option as source widget.
    The value of source widget changes upon change in value
    of target.

    Parameters
    ----------
    `source`: The widget to be converted.
    `target`: The type of widget to be generated.

    Example:
    ```
    discrete_slider = pn.widgets.DiscreteSlider(name='Discrete Slider',
                                                options=[2, 4, 8, 16, 32, 64],
                                                value=32)
    selector = convert_widget(discrete_slider,pn.widgets.Select())
    ```
    """
    target_w = target
    target_w.name = source.params.args[0].name
    target_w.options = source.params.args[0].options
    target_w.value = source.params.args[0].value

    def callback(*events):
        for event in events:
            if event.name == 'value':
                source.value = event.new

    target_w.param.watch(callback, ['value'], onlychanged=False)

    return target_w


def _is_coord(data, name):
    if name in list(data.coords):
        return name + " " + '\U0001F4C8'
    else:
        return name


def player_with_name_and_value(source):
    """
    source: pn.widgets.DiscretePlayer()
    target: consists of source player's name, value and player itself

    With pn.widgets.DiscretePlayer, we don't get name and
    value updates in textual form. This method is useful
    in case we want name and continuous value update.
    """
    mark = pn.pane.Markdown(f'{source.value}')

    def callback(*events):
        for event in events:
            if event.name == 'value':
                mark.object = str(event.new)

    source.param.watch(callback, ['value'], onlychanged=False)
    target = pn.Column(pn.Row(source.name, mark), source)
    return target


def is_float(a):
    # int, 'Nan', 'inf' and '-inf' are also float
    try:
        float(a)
        return True
    except:
        return False

if has_cartopy:
    def proj_params(proj):
        proj = getattr(ccrs, proj)
        params = inspect.getfullargspec(proj)
        out = {}
        if 'self' in params.args:
            params.args.remove('self')
        if len(params.args) and len(params.args) == len(params.defaults):
            for arg, val in zip(params.args, params.defaults):
                out.update({arg: val})
        return str(out)



import xarray as xr
import panel as pn
from xrviz.dashboard import Dashboard, find_cmap_limits
import pytest
from . import data
from ..utils import _is_coord
from ..compatibility import has_cartopy, has_crick_tdigest


@pytest.fixture(scope='module')
def dashboard(data):
    return Dashboard(data)


@pytest.fixture(scope='module')
def dash_for_Array(data):
    return Dashboard(data.temp)


def test_dashboard_with_invalid_datatype():
    with pytest.raises(ValueError):
        Dashboard(str)


def test_check_is_plottable_1D(dashboard):
    # `validtime` is 1D variable
    dashboard.plot_button.disabled = False
    dashboard.control.displayer.select_variable('validtime')
    assert dashboard.plot_button.disabled


def test_check_is_plottable_coord(dashboard):
    # `time` is data coordinate
    dashboard.plot_button.disabled = False
    dashboard.control.displayer.select_variable('time')
    assert dashboard.plot_button.disabled


def test_check_is_plottable_other_vars(dashboard):
    # `temp` is neither 1D nor coordinate
    dashboard.plot_button.disabled = True
    dashboard.control.displayer.select_variable('temp')
    assert dashboard.plot_button.disabled is False


def test_2d_variable_for_dims(dashboard):
    dashboard.control.displayer.select_variable('lat')
    fields = dashboard.control.fields
    assert fields.x.value == 'nx'
    assert fields.y.value == 'ny'
    assert [agg_sel.name for agg_sel in fields.agg_selectors] == []
    dashboard.plot_button.clicks += 1
    graph = dashboard.output[0]
    assert isinstance(graph, pn.pane.holoviews.HoloViews)
    assert [index_sel for index_sel in dashboard.output[1]] == []


def test_3d_variable_for_dims(dashboard):
    dashboard.control.displayer.select_variable('air_v')
    fields = dashboard.control.fields
    assert fields.x.value == 'nx'
    assert fields.y.value == 'ny'
    agg_selectors = [agg_sel.name for agg_sel in fields.agg_selectors]
    assert agg_selectors == ['time']
    dashboard.plot_button.clicks += 1
    graph = dashboard.output[0]
    assert isinstance(graph, pn.pane.holoviews.HoloViews)
    index_selectors = [index_sel.name for index_sel in dashboard.output[1]]
    assert index_selectors == ['time']


@pytest.mark.parametrize('dashboard',
                         ['dashboard', 'dash_for_Array'],
                         indirect=True)
def test_4d_variable_for_dims(dashboard):
    dashboard.control.displayer.select_variable('temp')
    fields = dashboard.control.fields
    assert fields.x.value == 'nx'
    assert fields.y.value == 'ny'
    agg_selectors = [agg_sel.name for agg_sel in fields.agg_selectors]
    assert agg_selectors == ['sigma', 'time']
    dashboard.plot_button.clicks += 1
    graph = dashboard.output[0]
    assert isinstance(graph, pn.pane.holoviews.HoloViews)
    index_selectors = [index_sel.name for index_sel in dashboard.output[1]]
    assert index_selectors == ['sigma', 'time']


def test_link_aggregation_selectors(dashboard):
    dashboard.control.displayer.select_variable('temp')
    llim = dashboard.control.style.lower_limit
    ulim = dashboard.control.style.upper_limit
    llim.value = '10'
    ulim.value = '20'
    agg_sel = dashboard.control.fields.agg_selectors[0]
    agg_sel.value = 'mean'
    assert llim.value is None
    assert ulim.value is None


def test_set_coords(dashboard):
    new_coords = sorted(['time', 'sigma', 'lat', 'lon'])
    dashboard.control.coord_setter.coord_selector.value = new_coords
    assert sorted(list(dashboard.data.coords)) == new_coords


def test_animate_wigdet_for_dims(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    fields = dashboard.control.fields
    agg_selectors = fields.agg_selectors
    agg_selectors[0].value = 'animate'
    dashboard.create_graph()
    assert isinstance(dashboard.output[0], pn.pane.holoviews.HoloViews)
    assert isinstance(dashboard.output[1][0][1], pn.widgets.player.DiscretePlayer)


@pytest.mark.skipif(not has_cartopy, reason='cartopy not present')
def test_with_is_geo_projection(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    dashboard.control.displayer.select_variable('temp')
    proj_panel = dashboard.control.projection
    proj_panel.is_geo.value = True
    proj_panel.projection.value = 'Orthographic'
    proj_panel.proj_params.value = "{'central_longitude': -78, 'central_latitude': 43, 'globe': None}"
    proj_panel.global_extent.value = True
    dashboard.create_graph()
    assert isinstance(dashboard.output[0], pn.pane.holoviews.HoloViews)
    assert isinstance(dashboard.output[1][0], pn.widgets.select.Select)


@pytest.mark.skipif(not has_cartopy, reason='cartopy not present')
def test_with_is_geo_basemap(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    dashboard.control.displayer.select_variable('temp')
    proj_panel = dashboard.control.projection
    proj_panel.is_geo.value = True
    dashboard.create_graph()
    assert isinstance(dashboard.output[0], pn.pane.holoviews.HoloViews)
    assert isinstance(dashboard.output[1][0], pn.widgets.select.Select)


def test_with_aggregations_for_dims(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    fields = dashboard.control.fields
    agg_selectors = fields.agg_selectors
    agg_selectors[0].value = 'max'
    agg_selectors[1].value = 'count'
    dashboard.plot_button.clicks += 1
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_with_aggregations_for_coords(dashboard):
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    agg_selectors = dashboard.control.fields.agg_selectors
    agg_selectors[0].value = 'max'
    agg_selectors[1].value = 'count'
    dashboard.plot_button.clicks += 1
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_color_scaling_for_dims(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    style = dashboard.control.style
    style.color_scale.value = 'log'
    dashboard.create_graph()
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_color_scaling_for_coords(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    dashboard.control.displayer.select_variable('temp')
    style = dashboard.control.style
    style.color_scale.value = 'log'
    dashboard.create_graph()
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_use_all_data_for_dims(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.style.use_all_data.value = True
    dashboard.create_graph()
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_use_all_data_for_coords(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.style.use_all_data.value = True
    dashboard.create_graph()
    assert isinstance(dashboard.output[0][0], pn.pane.holoviews.HoloViews)


def test_create_taps_and_series_graph_for_dims(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.fields.s_selector.value = 'sigma'
    dashboard.create_graph()
    dashboard.create_taps_graph(x=35, y=10)
    assert isinstance(dashboard.series_graph[0], pn.pane.holoviews.HoloViews)


def test_create_taps_and_series_graph_for_2d_coords(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.fields.s_selector.value = 'sigma'
    dashboard.create_graph()
    dashboard.create_taps_graph(x=-79.232, y=43.273)
    assert isinstance(dashboard.series_graph[0], pn.pane.holoviews.HoloViews)


@pytest.mark.skipif(not has_crick_tdigest, reason='crick.tdigest not present')
def test_find_cmap_limits_with_crick_tdigest():
    ds = xr.tutorial.open_dataset('air_temperature',
                                  chunks={'lat': 25, 'lon': 25, 'time': 10})
    a, b = find_cmap_limits(ds.air)
    assert abs(a - 255.38780056044027) < 0.1 and abs(b - 298.5900340551101) < 0.1



import pytest
import panel as pn
from xrviz.fields import Fields
from . import data
from .test_dashboard import dashboard
from ..compatibility import mpcalc


@pytest.fixture()
def fields(data):
    return Fields(data)


@pytest.fixture()
def fields_array(data):
    return Fields(data.temp)


@pytest.mark.parametrize('fields',
                         ['fields', 'fields_array'],
                         indirect=True)
def test_fields_initial(fields):
    assert fields.x.value is None
    assert fields.y.value is None
    assert fields.panel.name == 'Axes'
    assert fields.panel[0][0][0].object == '### Plot Dimensions'
    assert isinstance(fields.panel[0][1], pn.Spacer)
    assert fields.panel[0][2][0].object == '### Aggregations'


def test_x_y_var_without_coords(dashboard):
    dashboard.control.displayer.select_variable('temp')
    assert dashboard.control.fields.x.value == 'nx'
    assert dashboard.control.fields.y.value == 'ny'


def test_x_y_var_with_coords(dashboard):
    dashboard.control.displayer.select_variable('temp')
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    x_val, y_val = ('lon', 'lat') if mpcalc else ('lat', 'lon')
    assert dashboard.control.fields.x.value == x_val
    assert dashboard.control.fields.y.value == y_val


def test_change_y_dim_selectors(dashboard):
    dashboard.control.coord_setter.coord_selector.value = ['time', 'sigma']
    dashboard.control.displayer.select_variable('temp')
    fields = dashboard.control.fields
    fields.x.value = 'time'
    assert fields.y.value == 'nx'
    assert fields.agg_selectors[0].name == 'ny'
    assert fields.agg_selectors[1].name == 'sigma'


def test_check_are_var_coords(dashboard):
    fields = dashboard.control.fields
    fields.x.value = 'nx'
    fields.y.value = 'ny'
    assert fields.check_are_var_coords() is False
    dashboard.control.coord_setter.coord_selector.value = ['lat', 'lon']
    fields.x.value = 'lon'
    fields.y.value = 'lat'
    assert fields.check_are_var_coords() is True



import panel as pn
from .compatibility import logger
from .sigslot import SigSlot


class CoordSetter(SigSlot):
    """
    An input pane for choosing which variables are considered coordinates.

    It uses a `Cross Selector <https://panel.pyviz.org/reference/widgets/CrossSelector.html>`_
    to display a list of simple and coordinate variables.
    Simple variables (which are not data coordinates) are available on
    left side and default coordinates are available on right side.
    To set variables as coordinates, make selection on left side and click
    ``>>``. Similarly making selection on right side and clicking ``<<``
    will reset the coordinates. Other panes update themselves
    accordingly, in response to this change.
    """
    def __init__(self, data):
        super().__init__()
        self.data = data
        self.coord_selector = pn.widgets.CrossSelector(
            name='Set Coords',
            value=list(self.data.coords),
            options=list(self.data.variables)
        )

        self.panel = self.coord_selector

    def set_coords(self, data):
        """
        Called when the data attribute of the interface has change in variables considered as coordinates.
        """
        self.data = data
        logger.debug(self.coord_selector.value)
        self.coord_selector.value = list(self.data.coords)

    def setup_initial_values(self, init_params={}):
        """
        To set the variables, whose names have been passed, as coordinates.
        """
        if 'Set Coords' in init_params:
            self.coord_selector.value = init_params['Set Coords']



import pytest
from xrviz.control import Control
from . import data
from ..compatibility import has_cartopy


@pytest.fixture()
def control(data):
    return Control(data)


def test_control_initial(control):
    tabs = [tab.name for tab in control.tabs]
    if has_cartopy:
        assert tabs == ['Variables', 'Set Coords', 'Axes', 'Style',
                        'Projection']
    else:
        assert tabs == ['Variables', 'Set Coords', 'Axes', 'Style']



import panel as pn
from holoviews.plotting import list_cmaps
from .sigslot import SigSlot


class Style(SigSlot):
    """
    A pane to customise styling of the graphs.

    The following options are available in this pane:

        1. ``height`` (default `300`):
            To modify the height of main and series graph.
        2. ``width`` (default `700`):
            To modify the width of the main and series graph.
        3. ``cmap`` (default `Inferno`):
            To select a colormap for the main graph.
        4. ``color_scale`` (default `linear`):
            To scale the values to be plotted.
            The scaling options available are ``linear``, ``exp``, ``log``,
            ``reciprocal``, ``square`` and ``sqrt``. Here ``linear`` implies
            no scaling.
        5. ``cmap limits``:
            To change the colormap limits. User can fill these limits before
            plotting a variable. In case not filled by user, automatic filling
            of limits happen.

            - ``lower limit``: auto-filled value equals ``quantile(0.1)`` of values to be plotted.
            - ``upper limit``: auto-filled value equals ``quantile(0.9)`` of values to be plotted.

            In case of dask array, `dask.array.percentile <https://docs.dask.org/en/latest/array-api.html#dask.array.percentile>`_
            is use to compute the limits. ``tdigest`` method is used in case `crick <https://pypi.org/project/crick/>`_ is present.
            The value of limits is rounded off to 5 decimal places, for simplicity.

            Note that these values are filled with respect to color scaled
            values. Also these limits clear upon change in variable or
            color scaling.

        6. ``compute min/max from all data`` (default `False`):
            - ``True``: all values present in a data variable are used to compute upper and lower colormap limits.
            - ``False``: only values necessary to create first step/instance of graph are used.

            It is better to have its value `False` for larger datasets, to save
            computation time.
        7. ``colorbar`` (default `True`):
            Provides option to display/hide colorbar.
        8. ``rasterize`` (default `True`):
            Provides option to use `data shading <http://datashader.org/>`_ .
            It is better to have its value `True`, to get highly optimized
            rendering.
    """

    def __init__(self):
        super().__init__()
        self.height = pn.widgets.IntSlider(name='height', value=300, start=100,
                                           end=1200)
        self.width = pn.widgets.IntSlider(name='width', value=700, start=100,
                                          end=1200)
        self.cmap = pn.widgets.Select(name='cmap', value='Inferno',
                                      options=list_cmaps())
        self.colorbar = pn.widgets.Checkbox(name='colorbar', value=True)
        # colormap_limits
        self.lower_limit = pn.widgets.TextInput(name='cmap lower limit',
                                                width=140)
        self.upper_limit = pn.widgets.TextInput(name='cmap upper limit',
                                                width=140)
        self.use_all_data = pn.widgets.Checkbox(name='compute min/max from all data', value=False)

        scaling_ops = ['linear', 'exp', 'log', 'reciprocal', 'square', 'sqrt']
        self.color_scale = pn.widgets.Select(name='color_scale',
                                             value='linear',
                                             options=scaling_ops)
        self.rasterize = pn.widgets.Checkbox(name='rasterize', value=True)

        self._register(self.use_all_data, 'clear_cmap_limits')
        self._register(self.color_scale, 'clear_cmap_limits')
        self.connect('clear_cmap_limits', self.setup)

        self.panel = pn.Column(pn.Row(self.height, self.width),
                               pn.Row(self.cmap, self.color_scale),
                               pn.Row(self.lower_limit, self.upper_limit),
                               pn.Row(self.use_all_data, self.colorbar,
                                      self.rasterize),
                               name='Style')

    def setup(self, *args):
        #  Clears cmap limits
        self.lower_limit.value = None
        self.upper_limit.value = None

    def setup_initial_values(self, init_params={}):
        """
        To select initial values for the widgets in this pane.
        """
        for row in self.panel:
            for widget in row:
                if widget.name in init_params:
                    widget.value = init_params[widget.name]

    @property
    def kwargs(self):
        out = {widget.name: widget.value
               for row in self.panel for widget in row}
        return out



from .dashboard import Dashboard


def example(show=True):
    """
    Generates interface for the sample dataset.

    Parameters
    ----------
    show: bool (True)
        ``True``: A new browser tab will be opened, and the function will block until interrupted.
        ``False``: The Dashboard instance is returned without being executed.
    """
    from .sample_data import great_lakes
    initial_params = {'Variables': 'temp',  # Select Variable
                      'Set Coords': ['lat', 'lon'],  # Set Coords
                      # Fields
                      'x': 'lon',
                      'y': 'lat',
                      'sigma': 'animate',

                      # style
                      'height': 300,
                      'width': 650,
                      'colorbar': True,
                      'cmap': 'Viridis',

                      # projection
                      'is_geo': True,
                      'basemap': 'OSM',
                      'crs': 'PlateCarree',
                      'projection': 'Orthographic',
                      'crs params': "{'central_longitude': 0.0}",
                      'projection params': "{'central_longitude': -78, 'central_latitude': 43, 'globe': None}"
                      }
    dash = Dashboard(great_lakes, initial_params=initial_params)
    if show:
        dash.show()
    return dash



import pytest
import panel as pn
from xrviz.style import Style
from holoviews.plotting import list_cmaps


@pytest.fixture()
def style():
    return Style()


def test_height(style):
    h_widget = style.height
    assert isinstance(h_widget, pn.widgets.IntSlider)
    assert h_widget.name == 'height'
    assert h_widget.value == 300
    assert h_widget.start == 100
    assert h_widget.end == 1200


def test_width(style):
    w_widget = style.width
    assert isinstance(w_widget, pn.widgets.IntSlider)
    assert w_widget.name == 'width'
    assert w_widget.value == 700
    assert w_widget.start == 100
    assert w_widget.end == 1200


def test_cmap(style):
    cmap_widget = style.cmap
    assert isinstance(cmap_widget, pn.widgets.Select)
    assert cmap_widget.name == 'cmap'
    assert cmap_widget.value == 'Inferno'
    assert cmap_widget.options == list_cmaps()


def test_colorbar(style):
    cbar_widget = style.colorbar
    assert isinstance(cbar_widget, pn.widgets.Checkbox)
    assert cbar_widget.name == 'colorbar'
    assert cbar_widget.value is True


def test_lower_limit(style):
    l_lim_widget = style.lower_limit
    assert isinstance(l_lim_widget, pn.widgets.TextInput)
    assert l_lim_widget.name == 'cmap lower limit'
    assert l_lim_widget.width == 140


def test_upper_limit(style):
    u_lim_widget = style.upper_limit
    assert isinstance(u_lim_widget, pn.widgets.TextInput)
    assert u_lim_widget.name == 'cmap upper limit'
    assert u_lim_widget.width == 140


def test_use_all_data(style):
    use_data_widget = style.use_all_data
    assert isinstance(use_data_widget, pn.widgets.Checkbox)
    assert use_data_widget.name == 'compute min/max from all data'
    assert use_data_widget.value is False


def test_color_scale(style):
    cs_widget = style.color_scale
    assert isinstance(cs_widget, pn.widgets.Select)
    assert cs_widget.name == 'color_scale'
    assert cs_widget.value == 'linear'
    scaling_ops = ['linear', 'exp', 'log', 'reciprocal', 'square', 'sqrt']
    assert cs_widget.options == scaling_ops


def test_rasterize(style):
    raster_wid = style.rasterize
    assert isinstance(raster_wid, pn.widgets.Checkbox)
    assert raster_wid.name == 'rasterize'
    assert raster_wid.value is True


def test_setup(style):
    style.lower_limit.value = '123'
    style.upper_limit.value = '123'
    style._emit('clear_cmap_limits', '')
    assert style.lower_limit.value is None
    assert style.upper_limit.value is None



import panel as pn
import xarray as xr
import warnings
from .sigslot import SigSlot
from .utils import convert_widget
from .compatibility import mpcalc


class Fields(SigSlot):
    """
    To select the fields to plot along.

    This pane controls which array dimensions should be mapped,
    how additional dimensions should be handled, and which dimension
    series plots should be extracted along.

    Parameters
    ----------
    data: xarray.DataSet

    Attributes
    ----------
    x:
        To select which of the available dimensions/coordinates in the data is
        assigned to the plot’s x (horizontal) axis.
    y:
        To select which of the available dimensions/coordinates in the data is
        assigned to the plot’s y (vertical) axis.

    Remaining Dims:
        Any one of the following aggregations can be applied on each of
        remaining dimensions:
            1. ``select``: It creates a ``pn.widgets.Select``, to select the value of dimension, for which the graph would be displayed.
            2. ``animate``: It creates a ``panel.widgets.DiscretePlayer`` which helps to quickly iterate over all the values for a dimension.
            3. ``mean``: Creates plot along mean of the selected dimension.
            4. ``max``: Creates plot along maximum of the selected dimension.
            5. ``min``: Creates plot along minimum of the selected dimension.
            6. ``median``: Creates plot along median of the selected dimension.
            7. ``std``: Creates plot along standard deviation of the selected dimension.
            8. ``count``: Creates plot along non-nan values of the selected dimension.

        Note that for both ``select`` and ``animate``, the plot will update
        according to the value selected in the generated widget. Also, if a
        dimension has been aggregated, its select widget would not be
        available.

    Extract Along:
        This selector provides the option to select the dimension along which to
        create a series graph.
    """

    def __init__(self, data):
        super().__init__()
        self.data = data

        dim_header = pn.pane.Markdown('### Plot Dimensions', margin=(0, 20, 0, 20))
        self.x = pn.widgets.Select(name='x', width=240, margin=(0, 20, 5, 20))
        self.y = pn.widgets.Select(name='y', width=240, margin=(0, 20, 20, 20))

        agg_header = pn.pane.Markdown('### Aggregations', margin=(0, 20, 0, 20))
        self.agg_selectors = pn.Column()
        self.agg_opts = ['select', 'animate', 'mean', 'max',
                         'min', 'median', 'std', 'count']
        self.series_col = pn.Column()
        self.are_var_coords = False

        self._register(self.x, 'x')
        self._register(self.y, 'y')

        self.connect('x', self.change_y)
        self.connect('y', self.change_dim_selectors)

        self.panel = pn.Column(
            pn.Row(
                pn.WidgetBox(dim_header, self.x, self.y,
                             background=(240, 240, 240)),
                pn.Spacer(),
                pn.WidgetBox(agg_header, self.agg_selectors,
                             background=(240, 240, 240))
            ),
            self.series_col,
            name='Axes')

    def setup(self, var):
        """
        Fill available options for the selected variable.
        """
        self.agg_selectors.clear()  # To empty previouly selected value from selector
        self.var = var if isinstance(var, str) else var[0]
        self.var_dims = list(self.data[var].dims)
        self.indexed_coords = set(self.var_dims).intersection(set(self.data[var].coords))
        self.non_indexed_coords = set(self.data[var].coords) - self.indexed_coords
        self.sel_options = sorted(self.var_dims + list(self.non_indexed_coords))

        self.x_guess, self.y_guess = self.guess_x_y(self.var) if mpcalc else ['None', 'None']
        x_opts = self.sel_options.copy()
        if len(x_opts):  # to check that data has dim (is not Empty)
            self.x.options = x_opts
            self.x.value = self.x_guess if self.x_guess and self.x_guess in x_opts else x_opts[0]
            y_opts = x_opts.copy()
            del y_opts[0]
            if y_opts is None:
                self.y.options = []
                self.remaining_dims = []
            else:
                self.remaining_dims = [opt for opt in y_opts
                                       if opt != self.y.value]
                self.change_y()

    def change_y(self, value=None):
        """
        Updates ``y`` by removing the value of ``x``, from the available options
        """
        # if x belong to var_dims replace the y with remaining var_dims
        # else if x belong to non_indexed_coords, replace y with remaining
        # non_indexed_coords
        values = self.sel_options.copy()
        x_val = self.x.value
        values.remove(x_val)

        if x_val in self.var_dims:
            valid_values = set(values) - self.non_indexed_coords
        else:  # x_val belong to non_indexed_coords
            values = set(values) - set(self.var_dims)
            #  Plot can be generated for 2 values only if ndims of both match
            valid_values = [val
                            for val in values if self.ndim_matches(x_val, val)]
        y_opts = sorted(list(valid_values))
        self.y.options = y_opts
        if len(y_opts):
            self.y.value = self.y_guess if self.y_guess and self.y_guess in y_opts else y_opts[0]
        self.are_var_coords = self.check_are_var_coords()
        self.change_dim_selectors()

    def change_dim_selectors(self, *args):
        """
        Updates the dimensions available for `Aggregation` and `Extract Along` upon change in value of ``y``.
        """
        self.are_var_coords = self.check_are_var_coords()
        self.agg_selectors.clear()
        self.series_col.clear()
        x = self.x.value
        y = self.y.value
        used_opts = {x, y}

        if x in self.var_dims:
            self.remaining_dims = [dim for dim in self.var_dims
                                   if dim not in used_opts]
        else:  # is a coord
            #  We can't aggregate along dims which are present in x and y.
            x_val_dims = set(self.data[x].dims) if x is not None else set()
            y_val_dims = set(self.data[y].dims) if y is not None else set()
            dims_not_to_agg = x_val_dims.union(y_val_dims).union(used_opts)
            self.remaining_dims = [dim for dim in self.var_dims
                                   if dim not in dims_not_to_agg]

        for i, dim in enumerate(sorted(self.remaining_dims)):
            if i == 0 and i == (len(self.remaining_dims)-1):
                margin = (0, 20, 20, 20)
            elif i == 0:
                margin = (0, 20, 5, 20)
            elif i == (len(self.remaining_dims)-1):
                margin = (5, 20, 20, 20)
            else:
                margin = (0, 20, 5, 20)
            agg_selector = pn.widgets.Select(name=dim,
                                             options=self.agg_opts,
                                             width=240, margin=margin)
            self._register(agg_selector, agg_selector.name)
            self.agg_selectors.append(agg_selector)

        self.s_selector = pn.widgets.Select(name='extract along',
                                            options=[None]+sorted(self.remaining_dims),
                                            width=240)
        self._register(self.s_selector, 'extract_along')
        self.series_col.append(self.s_selector)

    def setup_initial_values(self, init_params={}):
        for widget in [self.x, self.y] + list(self.agg_selectors) + list(self.series_col):
            if widget.name in init_params:
                widget.value = init_params[widget.name]

    @property
    def kwargs(self):
        # Column(name='Axes')
        #     [0] Row
        #         [0] Column(background='rgb(175,175,175)')
        #             [0] Markdown(str)
        #             [1] Select(name='x', width=200)
        #             [2] Select(name='y', width=200)
        #         [1] Spacer()
        #         [2] Column(background='rgb(175,175,175)')
        #             [0] Markdown(str)
        #             [1] Column()
        #     [1] Column
        #         [0] Select()
        out = {p.name: p.value for p in self.panel[0][0][1:]}  # since panel[0][0][1] is Markdown
        selectors = {p.name: p.value for p in self.panel[0][2][1]}  # remaining_dims
        out.update(selectors)
        dims_to_select_animate = [dim for dim, agg in selectors.items() if agg in ['select', 'animate']]
        dims_to_agg = [dim for dim in selectors
                       if dim not in dims_to_select_animate]
        out.update({'dims_to_agg': dims_to_agg})
        out.update({'dims_to_select_animate': sorted(dims_to_select_animate)})
        out.update({'are_var_coords': self.are_var_coords})
        # remaining_dims = dims_to_agg + dims_to_select_animate
        out.update({'remaining_dims': self.remaining_dims})
        out.update({p.name: p.value for p in self.series_col})
        return out

    def set_coords(self, data, var):
        self.data = data
        if var:
            self.setup(var)

    def ndim_matches(self, var1, var2):
        return self.data[var1].ndim == self.data[var2].ndim

    def check_are_var_coords(self):
        """
        Check if both ``x`` and ``y`` are coordinates of the selected variable.
        """
        var_coords = list(self.data[self.var].coords)
        x = self.x.value
        y = self.y.value
        return True if x in var_coords and y in var_coords else False

    def guess_x_y(self, var):
        """
        To guess the value of ``x`` and ``y`` with `metpy.parse_cf`_.
        This is applicable only for the case when both `x` and `y` are data
        coordinates.

        .. _`metpy.parse_cf`: https://github.com/Unidata/MetPy/blob/master/metpy/xarray.py#L335
        """
        try:
            parsed_var = self.data.metpy.parse_cf(var)
            x, y = parsed_var.metpy.coordinates('x', 'y')
            return [coord.name for coord in (x, y)]
        except:  # fails when coords have not been set or available.
            return [None, None]



import panel as pn
import xarray as xr
from .sigslot import SigSlot
from .utils import _is_coord


class Display(SigSlot):
    """Displays a list of data variables for selection.

    For each data, its variables are displayed in ``pn.widgets.MultiSelect``.
    Variables which are data coordinates are annotated with '📈'.

    Parameters
    ----------

    data: xarray.DataSet

    Attributes
    ----------

    panel:
        A ``panel.Row`` instance which displays the Multiselect widget.

    """

    def __init__(self, data):
        super().__init__()
        self.data = data
        self.select = pn.widgets.MultiSelect(size=8, max_width=300,
                                             height=210,
                                             width_policy='max',
                                             name='Variables')
        self.set_variables()

        self._register(self.select, "variable_selected")

        self.panel = pn.Row(self.select)

    def set_variables(self,):
        self.select.options = {
            _is_coord(self.data, name): name
            for name in list(self.data.variables)
        }

    def select_variable(self, variable):
        """
        Select a data variable from the available options.
        """
        if isinstance(variable, str):
            if variable in self.select.options.values():
                self.select.value = [variable]
            else:
                print(f"Variable {variable} not present in displayer.")

    def setup_initial_values(self, init_params={}):
        if 'Variables' in init_params:
            self.select_variable(init_params['Variables'])

    @property
    def kwargs(self):
        # Select only the first value from the selected variables.
        out = {p.name: p.value[0] for p in self.panel}
        return out

    def set_coords(self, data):
        self.data = data
        self.set_variables()



import ast
import dask
import panel as pn
import pandas as pd
import numpy as np
import xarray as xr
import hvplot.xarray
import hvplot.pandas
import holoviews as hv
from holoviews import streams
from bokeh.models import HoverTool
import warnings
from itertools import cycle
import numpy
from .sigslot import SigSlot
from .control import Control
from .utils import convert_widget, player_with_name_and_value, is_float
from .compatibility import ccrs, gv, gf, has_cartopy, logger, has_crick_tdigest


class Dashboard(SigSlot):
    """
    Main entry point to XrViz, an interactive GUI for a given dataset.

    Parameters
    ----------
    data: xarray.DataSet
        The data to be visualised

    initial_params: `dict`
        To pre-select values of widgets upon initialization. The keys are
        generally names of widgets within the input area of the interface.
        For more details, refer to
        `Set Initial Parameters <../html/set_initial_parameters.html>`_ .

    Attributes
    ----------

    1. panel:
            A ``panel.Tabs`` instance containing the user input panes and
            output graphs of the interface.
    2. control:
            A ``Control`` instance responsible for input panes (control panel).
    3. plot_button:
            A ``pn.widgets.Button`` that generates graph according to values
            selected in input panes, upon click.
    4. graph:
            A ``HoloViews(DynamicMap)`` instance containing the main graph.
    5. output:
            The ``graph`` along with the select widgets for index selection.
    6. taps_graph:
            A ``holoviews.Points`` instance to record the location of taps.
    7. series_graph:
            A ``HoloViews(Overlay)`` instance having series extracted.
    8. clear_series_button:
            A ``pn.widgets.Button`` to clear the `taps_graph` and
            `series_graph`.
    """
    def __init__(self, data, initial_params={}):
        super().__init__()
        if not isinstance(data, xr.core.dataarray.DataWithCoords):
            raise ValueError("Input should be an xarray data object, not %s" % type(data))
        self.set_data(data)
        self.initial_params = initial_params
        self.control = Control(self.data)
        self.plot_button = pn.widgets.Button(name='Plot', width=200,
                                             disabled=True)
        self.index_selectors = []
        self.graph = pn.Spacer(name='Graph')
        self.taps_graph = hv.Points([])
        self.series_graph = pn.Row(pn.Spacer(name='Series Graph'))
        self.clear_series_button = pn.widgets.Button(name='Clear',
                                                     width=200,
                                                     disabled=True)
        self.output = pn.Row(self.graph,
                             pn.Column(name='Index_selectors'))

        self._register(self.plot_button, 'plot_clicked', 'clicks')
        self.connect('plot_clicked', self.create_graph)

        self._register(self.control.coord_setter.coord_selector, 'set_coords')
        self.connect("set_coords", self.set_coords)

        self._register(self.clear_series_button, 'clear_series', 'clicks')
        self.connect('clear_series', self.clear_series)

        self.control.displayer.connect('variable_selected',
                                       self.check_is_plottable)
        self.control.displayer.connect('variable_selected',
                                       self._link_aggregation_selectors)
        self.control.fields.connect('x', self._link_aggregation_selectors)
        self.control.fields.connect('y', self._link_aggregation_selectors)

        self.panel = pn.Column(self.control.panel,
                               pn.Row(self.plot_button,
                                      self.clear_series_button),
                               self.output,
                               self.series_graph, width_policy='max')

        # To auto-select in case of single variable
        if len(list(self.data.variables)) == 1:
            self.control.displayer.select.value = list(self.data.variables)

        self.control.setup_initial_values(self.initial_params)
        self.taps = []
        self.tap_stream = streams.Tap(transient=True)
        colors = ['#60fffc', '#6da252', '#ff60d4', '#ff9400', '#f4e322',
                  '#229cf4', '#af9862', '#629baf', '#7eed5a', '#e29ec8',
                  '#ff4300']
        self.color_pool = cycle(colors)
        self.clear_points = hv.streams.Stream.define(
            'Clear_points', clear=False)(transient=True)

    def clear_series(self, *args):
        """
        Clears the markers on the image, and the extracted series.
        """
        if not self.clear_series_button.disabled:
            self.series_graph[0] = pn.Spacer(name='Series Graph')
            self.series = hv.Points([]).opts(height=self.kwargs['height'],
                                             width=self.kwargs['width'])
            self.taps.clear()
            self.clear_points.event(clear=True)

    def _link_aggregation_selectors(self, *args):
        for dim_selector in self.control.kwargs['remaining_dims']:
            self.control.fields.connect(dim_selector, self.control.style.setup)

    def create_graph(self, *args):
        """
        Creates a graph according to the values selected in the widgets.

        This method is usually invoked by the user clicking "Plot"

        It handles the following two cases:

            1. Both `x`, `y` are present in selected variable's coordinates.
            Geographic projection is possible only in this case. It uses
            ``create_selectors_players`` method for creation of the graph.
            Here the selectors generated automatically by hvplot are used.

            2. One or both of  `x`, `y` are NOT present in selected variable's
            coordinates (both `x` and `y` are considered as dimensions). It
            uses ``create_indexed_graph`` method for creation of the graph.
            The selectors are created and linked with graph by XrViz.
        """
        self.kwargs = self.control.kwargs
        self.var = self.kwargs['Variables']
        if self.index_selectors:
            for selector in self.index_selectors:
                del selector
        self.index_selectors = []
        self.output[1].clear()  # clears Index_selectors
        self.series_graph[0] = pn.Spacer(name='Series Graph')
        self.series = hv.Points([]).opts(height=self.kwargs['height'],
                                         width=self.kwargs['width'])
        self.taps.clear()
        self.control.fields.connect('extract_along', self.clear_series)

        are_var_coords = self.kwargs['are_var_coords']
        if are_var_coords:
            graph_opts = {'x': self.kwargs['x'],
                          'y': self.kwargs['y'],
                          'title': self.var,
                          'height': self.kwargs['height'],
                          'width': self.kwargs['width'],
                          'cmap': self.kwargs['cmap'],
                          'colorbar': self.kwargs['colorbar'],
                          'rasterize': self.kwargs['rasterize']}
            color_scale = self.kwargs['color_scale']
            dims_to_agg = self.kwargs['dims_to_agg']
            use_all_data = self.kwargs['compute min/max from all data']
            sel_data = self.data[self.var]

            if has_cartopy:
                is_geo = self.kwargs['is_geo']
                base_map = self.kwargs['basemap']
                show_map = True if base_map != None else False

                if is_geo:
                    crs_params = self.kwargs['crs params']
                    crs_params = process_proj_params(crs_params)
                    crs = getattr(ccrs, self.kwargs['crs'])(**crs_params)
                    geo_ops = {'alpha': self.kwargs['alpha'],
                               'project': self.kwargs['project'],
                               'global_extent': self.kwargs['global_extent'],
                               'geo': True,
                               'crs': crs}
                    if not show_map:
                        # find projection and crs, add it to geo_ops
                        proj_val = self.kwargs['projection']
                        if proj_val:
                            proj_params = self.kwargs['projection params']
                            proj_params = process_proj_params(proj_params)
                            projection = getattr(ccrs, self.kwargs['projection'])(**proj_params)
                            geo_ops.update({'projection': projection})

                    graph_opts.update(geo_ops)

                    feature_map = gv.Overlay([getattr(gf, feat) for feat in self.kwargs['features'] if feat is not 'None'])

            for dim in dims_to_agg:
                if self.kwargs[dim] == 'count':
                    sel_data = (~ sel_data.isnull()).sum(dim)
                else:
                    agg = self.kwargs[dim]
                    sel_data = getattr(sel_data, agg)(dim)

            if self.var in list(sel_data.coords):  # When a var(coord) is plotted wrt itself
                sel_data = sel_data.to_dataset(name=f'{sel_data.name}_')

            if color_scale is not 'linear':
                sel_data = getattr(numpy, color_scale)(sel_data)  # Color Scaling

            if not use_all_data:
                # sel the values at first step, to use for cmap limits
                sels = {dim: 0 for dim in self.kwargs['dims_to_select_animate']}
                sel_data_for_cmap = sel_data.isel(**sels, drop=True)
            else:
                sel_data_for_cmap = sel_data

            cmin, cmax = self.kwargs['cmap lower limit'], self.kwargs['cmap upper limit']
            cmin, cmax = (cmin, cmax) if is_float(cmin) and is_float(cmax) else ('', '')

            # It is better to set initial values as 0.1,0.9 rather than
            # 0,1(min, max) to get a color balance graph
            c_lim_lower, c_lim_upper = (
                (float(cmin), float(cmax)) if cmin and cmax
                else find_cmap_limits(sel_data_for_cmap))

            color_range = {sel_data.name: (c_lim_lower, c_lim_upper)}

            if not cmin:  # if user left blank or initial values are empty
                self.control.style.lower_limit.value = str(round(c_lim_lower, 5))
                self.control.style.upper_limit.value = str(round(c_lim_upper, 5))

            assign_opts = {dim: self.data[dim] for dim in sel_data.dims}
            # Following tasks are happening here:
            # 1. assign_opts: reassignment of coords(if not done result in
            # errors for some of the selections in fields panel)
            # 2. graph_opts: customise the plot according to selections in
            # style and projection(if available)
            # 3. color_range: customise the colormap range according to cmap
            # lower and upper limits
            # 4. active_tools: activate the tools required such as 'wheel_zoom',
            # 'pan'
            graph = sel_data.assign_coords(
                **assign_opts).hvplot.quadmesh(
                **graph_opts).redim.range(**color_range).opts(
                active_tools=['wheel_zoom', 'pan'])

            self.tap_stream.source = graph

            if has_cartopy and is_geo:
                graph = (
                    feature_map * graph
                    if self.kwargs['features'] != ['None'] else graph
                )
                if show_map:
                    graph = base_map * graph
                    self.tap_stream.source = graph

            self.create_selectors_players(graph)

        else:  # if one or both x,y are var_dims
            self.var_dims = list(self.data[self.var].dims)
            #  var_selector_dims refers to dims for which index_selectors
            #  would be created
            self.var_selector_dims = self.kwargs['dims_to_select_animate']

            for dim in self.var_selector_dims:
                ops = list(self.data[self.var][dim].values)

                if self.kwargs[dim] == 'select':
                    selector = pn.widgets.Select(name=dim, options=ops)
                else:
                    selector = pn.widgets.DiscretePlayer(name=dim,
                                                         value=ops[0],
                                                         options=ops)
                self.index_selectors.append(selector)
                self._register(selector, selector.name)
                self.connect(selector.name, self.create_indexed_graph)

            self.create_indexed_graph()
            for selector in self.index_selectors:
                if isinstance(selector, pn.widgets.Select):
                    self.output[1].append(selector)
                else:
                    player = player_with_name_and_value(selector)
                    self.output[1].append(player)

    def create_indexed_graph(self, *args):
        """
        Creates a graph for the dimensions selected in widgets `x` and `y`.

        This is used when values selected in `x` and `y` are not data
        coordinates (i.e. one or both values are data dimensions).
        """
        self.kwargs = self.control.kwargs
        selection = {}  # to collect the value of index selectors
        for i, dim in enumerate(list(self.var_selector_dims)):
            selection[dim] = self.index_selectors[i].value
        graph_opts = {'x': self.kwargs['x'],
                      'y': self.kwargs['y'],
                      'title': self.var,
                      'height': self.kwargs['height'],
                      'width': self.kwargs['width'],
                      'cmap': self.kwargs['cmap'],
                      'colorbar': self.kwargs['colorbar'],
                      'rasterize': self.kwargs['rasterize']}
        dims_to_agg = self.kwargs['dims_to_agg']
        color_scale = self.kwargs['color_scale']
        use_all_data = self.kwargs['compute min/max from all data']

        sel_data = self.data[self.var]

        for dim in dims_to_agg:
            if self.kwargs[dim] == 'count':
                sel_data = (~ sel_data.isnull()).sum(dim)
            else:
                agg = self.kwargs[dim]
                sel_data = getattr(sel_data, agg)(dim)

        # rename the sel_data in case it is a coordinate, because we
        # cannot create a Dataset from a DataArray with the same name
        #  as one of its coordinates
        if sel_data.name in self.data.coords:
                sel_data = sel_data.to_dataset(name=f'{sel_data.name}_')

        if not use_all_data:  # do the selection earlier
            sel_data = sel_data.sel(**selection, drop=True)

        if color_scale is not 'linear':
            sel_data = getattr(numpy, color_scale)(sel_data)  # Color Scaling

        cmin, cmax = self.kwargs['cmap lower limit'], self.kwargs['cmap upper limit']
        cmin, cmax = (cmin, cmax) if is_float(cmin) and is_float(cmax) else ('', '')

        # It is better to set initial values as 0.1,0.9 rather than
        # 0,1(min, max) to get a color balance graph
        c_lim_lower, c_lim_upper = (
            (float(cmin), float(cmax)) if cmin and cmax
            else find_cmap_limits(sel_data))

        color_range = {sel_data.name: (c_lim_lower, c_lim_upper)}

        if not cmin:  # if user left blank or initial values are empty
            self.control.style.lower_limit.value = str(round(c_lim_lower, 5))
            self.control.style.upper_limit.value = str(round(c_lim_upper, 5))

        if use_all_data:  # do the selection later
            sel_data = sel_data.sel(**selection, drop=True)

        assign_opts = {dim: self.data[dim] for dim in sel_data.dims}
        graph = sel_data.assign_coords(
            **assign_opts).hvplot.quadmesh(**graph_opts).redim.range(
            **color_range).opts(active_tools=['wheel_zoom', 'pan'])
        self.graph = graph
        if len(self.data[self.var].dims) > 2 and self.kwargs['extract along']:
            self.tap_stream.source = graph
            self.taps_graph = hv.DynamicMap(
                self.create_taps_graph,
                streams=[self.tap_stream, self.clear_points])
            self.output[0] = self.graph * self.taps_graph
            self.clear_series_button.disabled = False
        else:
            self.output[0] = self.graph
            self.clear_series_button.disabled = True

    def create_taps_graph(self, x, y, clear=False):
        """
        Create an output layer in the graph which responds to taps

        Whenever the user taps (or clicks) the graph, a glyph will be overlaid,
        and a series is extracted at that point.
        """
        color = next(iter(self.color_pool))
        if None not in [x, y]:
            self.taps.append((x, y, color))
        if self.control.kwargs['extract along'] is None:
            self.taps = []
        is_geo = self.kwargs['is_geo'] if 'is_geo' in self.kwargs else None
        geo_disabled = self.control.projection.is_geo.disabled if is_geo else None

        # Choose between gv.Points and hv.Points
        if is_geo and geo_disabled is False:
            tapped_map = gv.Points(self.taps, vdims=['z'])
        else:
            tapped_map = hv.Points(self.taps, vdims=['z'])
        tapped_map.opts(color='z', marker='triangle', line_color='black',
                        size=8)
        self.series_graph[0] = self.create_series_graph(x, y, color, clear)
        return tapped_map

    def create_series_graph(self, x, y, color, clear=False):
        """
        Extract a series at a given point, and plot it.

        The series plotted has same color as that of the marker depicting the
        location of the tap.

        The following cases have been handled:
            `Case 1`:
                When both x and y are NOT coords (i.e. are dims)

            `Case 2`:
                When both x and y are coords

                ``2a``: Both are 1-dimensional

                ``2b``: Both are 2-dimensional with same dimensions.

                ``2c``: Both are 2-dimensional with different dims or are multi-dimcoordinates. Here we are unable to extract.
            Note that ``Case 1`` and ``Case 2a`` can be handled with the same
            code.
        """
        extract_along = self.control.kwargs['extract along']
        if None not in [x, y] and extract_along:
            color = self.taps[-1][-1] if self.taps[-1][-1] else None
            other_dims = [dim for dim in self.kwargs['remaining_dims'] if
                          dim is not extract_along]

            # to use the value selected in index selector for selecting
            # data to create series. In case of aggregation, plot is
            # created along 0th val of the dim.
            if len(other_dims):
                other_dim_sels = {}
                for dim in other_dims:
                    dim_found = False
                    for dim_sel in self.index_selectors:
                        long_name = self.data[dim].long_name if hasattr(
                            self.data[dim], 'long_name') else None
                        if dim_sel.name == dim or dim_sel.name == long_name:
                            val = dim_sel.value
                            other_dim_sels.update({dim: val})
                            dim_found = True
                    if not dim_found:  # when dim is used for aggregation
                        val = self.data[dim][0].values
                        other_dim_sels.update({dim: val})

            # Case 1 and  2a
            if not self.kwargs['are_var_coords'] or self.both_coords_1d():
                series_sel = {
                    self.kwargs['x']: self.correct_val(self.kwargs['x'], x),
                    self.kwargs['y']: self.correct_val(self.kwargs['y'], y)}
            # Case 2b
            elif self.both_coords_2d_with_same_dims():
                y_dim, x_dim = self.data[self.kwargs['x']].dims

                y_mean = self.data[self.kwargs['y']].mean() * np.pi / 180.
                a = (self.data[self.kwargs['y']] - y) ** 2 + (
                            (self.data[self.kwargs['x']] - x) * np.cos(
                        y_mean)) ** 2
                j, i = np.unravel_index(a.argmin(), a.shape)

                series_sel = {x_dim: self.correct_val(x_dim, i),
                              y_dim: self.correct_val(y_dim, j)}
            # Case 2c
            else:
                logger.debug("Cannot extract 2d coords with different dims and"
                             " multi-dimensional coords.")
                return self.series

            if len(other_dims):
                series_sel.update(other_dim_sels)

            sel_series_data = self.data[self.var]
            for dim, val in series_sel.items():
                sel_series_data = sel_val_from_dim(sel_series_data, dim, val)

            series_df = pd.DataFrame({extract_along: self.data[extract_along],
                                      self.var: np.asarray(sel_series_data)})

            tooltips = [(extract_along, f"@{extract_along}"),
                        (self.var, f"@{self.var}")]
            if len(other_dims):
                for dim, val in other_dim_sels.items():
                    tooltips.append((dim, str(val)))
            hover = HoverTool(tooltips=tooltips)

            series_map = series_df.hvplot(x=extract_along, y=self.var,
                                          height=self.kwargs['height'],
                                          width=self.kwargs['width'],
                                          tools=[hover])
            self.series = series_map.opts(color=color) * self.series

        return self.series

    def create_selectors_players(self, graph):
        """
        Converts the sliders generated by hvplot into selectors/players.

        This is applicable only when both `x` and `y` are present in variable
        coordinates. It converts any sliders generated by hvplot into
        selectors/players and moves them to the bottom of graph.
        """
        if len(self.data[self.var].dims) > 2 and self.kwargs['extract along']:
            self.taps_graph = hv.DynamicMap(self.create_taps_graph,
                                            streams=[self.tap_stream,
                                                     self.clear_points])
            self.clear_series_button.disabled = False
            graph = graph * self.taps_graph
        else:
            self.clear_series_button.disabled = True
        graph = pn.Row(graph)
        try:
            if graph[0][1]:  # if sliders are generated
                self.output[0] = graph[0][0]

                # link the generated slider with agg selector in fields
                for slider in graph[0][1]:
                    for dim in self.kwargs['dims_to_select_animate']:
                        long_name = self.data[dim].long_name if hasattr(
                            self.data[dim], 'long_name') else None
                        if slider.name == dim or slider.name == long_name:
                            if self.kwargs[dim] == 'select':
                                selector = convert_widget(slider,
                                                          pn.widgets.Select())
                            else:
                                selector = convert_widget(
                                    slider, pn.widgets.DiscretePlayer())
                            self.index_selectors.append(selector)

                for selector in self.index_selectors:
                    if isinstance(selector, pn.widgets.Select):
                        self.output[1].append(selector)
                    else:
                        player = player_with_name_and_value(selector)
                        self.output[1].append(player)

        except:  # else return simple graph
            self.output[0] = graph

    def set_data(self, data):
        self.data = (
            xr.Dataset({f'{data.name}': data})
            if isinstance(data, xr.DataArray) else data
        )

    def set_coords(self, *args):
        # We can't reset indexed coordinates so add them every time
        # in coord_selector.value
        self.data = self.data.reset_coords()
        indexed_coords = set(self.data.dims).intersection(set(self.data.coords))
        new_coords = set(args[0]).union(indexed_coords)
        self.data = self.data.set_coords(new_coords)  # this `set_coords` belongs to xr.dataset
        self.control.set_coords(self.data)

    def check_is_plottable(self, var):
        """
        Check if a data variable can be plotted.

        If a variable is 1-d, disable plot_button for it.
        """
        self.plot_button.disabled = False  # important to enable button once disabled
        data = self.data[var[0]]
        self.plot_button.disabled = len(data.dims) <= 1

    def correct_val(self, dim, x):
        """ Convert tapped coordinates to int, if not time-type
        """
        dtype = self.data[dim].dtype.kind
        if dtype == 'i':
            return int(x)
        elif dtype == 'f':
            return float(x)
        else:
            return str(x)

    def both_coords_1d(self):
        return len(self.data[self.kwargs['x']].dims) == 1 and len(self.data[self.kwargs['y']].dims) == 1

    def both_coords_2d_with_same_dims(self):
        x_dims = self.data[self.kwargs['x']].dims
        y_dims = self.data[self.kwargs['y']].dims
        return len(x_dims) == len(y_dims) == 2 and sorted(x_dims) == sorted(y_dims)


def find_cmap_limits(sel_data):
    if isinstance(sel_data.data, dask.array.core.Array):
        method = 'tdigest' if has_crick_tdigest else 'default'
        return dask.array.percentile(sel_data.data.ravel(), (10, 90),
                                     method=method).compute()
    else:  # if sel_data.data is numpy.ndarray
        return [float(q) for q in sel_data.quantile([0.1, 0.9])]


def sel_val_from_dim(data, dim, x):
    """ Select values from a dim.
    """
    try:
        return data.sel({dim: x})
    except:
        return data.sel({dim: x}, method='nearest')


def process_proj_params(params):
    params = ast.literal_eval(params)
    for k, v in params.items():
        if k == 'globe' and params['globe']:
            globe = ccrs.Globe(**v)
            params.update({'globe': globe})
    return params



from jinja2 import Environment, FileSystemLoader
import panel as pn
import os
import sys
import xarray as xr
from .sigslot import SigSlot


class Describe(SigSlot):
    """
    Displays the properties of the variable selected in the Display section.

    This section has two tables as output. The first table shows
    `Variable attributes` and the second table shows `Global Attributes`.
    Upon selection of a new variable in the Display widget, the first table
    updates itself with properties of the new selection,
    while the second table stays same.

    Parameters
    ----------

    data: xarray.DataSet

    Attributes
    ----------

    panel:
        A ``panel.pane.HTML`` instance which displays the tables arranged
        next to each other.

    """
    def __init__(self, data):
        super().__init__()
        self.data = data
        self.panel = pn.pane.HTML(
            style={'font-size': '12pt'}, min_width=800, height=100) 
        self.panel.object = "Description Section"
        self._template_load_path = os.path.join(os.path.dirname(__file__),
                                                "templates")
        self._template_env = Environment(loader=FileSystemLoader(self._template_load_path))
        self._variable_template = self._template_env.get_template('variable.html')

    def setup(self, var):
        self.panel.object = self.variable_pane_for_dataset(var)

    def variable_pane_for_dataset(self, var):
        """
        Returns HTML template describing variable and global attributes.
        """
        if var is not None:
            var = var if isinstance(var, str) else var[0]
            var_attrs = [(k, v) for k, v in self.data[var].attrs.items()]
            var_coords = [coord for coord in self.data[var].coords]
            var_dims = self.data[var].dims
            var_dtype = str(self.data[var].dtype)
            var_name = self.data[var].name
            var_nbytes = self.data[var].nbytes
            var_shape = self.data[var].shape
            var_size = self.data[var].size
            var_dim_shape = [(dim, shape) for dim, shape in zip(var_dims, var_shape)]
            data_attrs = [(k, v) for k, v in self.data.attrs.items()]
            data_coords = [coord for coord in self.data.coords.keys()]
            data_dims = [(k, v) for k, v in self.data.dims.items()]

            return self._variable_template.render(data_attrs=data_attrs,
                                                  data_coords=data_coords,
                                                  data_dims=data_dims,
                                                  var=var,
                                                  var_attrs=var_attrs,
                                                  var_coords=var_coords,
                                                  var_dim_shape=var_dim_shape,
                                                  var_dtype=var_dtype,
                                                  var_name=var_name,
                                                  var_nbytes=var_nbytes,
                                                  var_size=var_size,
                                                  )
        else:
            return self._variable_template.render(var=None)

    def set_coords(self, data, var):
        self.data = data
        if var is not None:
            self.setup(var=[var])



import inspect
import panel as pn
import xarray as xr
import geoviews as gv
from geoviews import tile_sources as gvts
from cartopy import crs as ccrs
from .sigslot import SigSlot
from .utils import proj_params

projections = [p for p in dir(ccrs) if inspect.isclass(getattr(ccrs, p)) and
               issubclass(getattr(ccrs, p), ccrs.Projection) and
               not p.startswith('_')]
not_to_include = ['Projection', 'UTM']
projections_list = sorted([p for p in projections if p not in not_to_include])


class Projection(SigSlot):
    """
    A pane to customise the projection of geographical data.

    The following options are available in this pane:

        1. ``is_geo`` (default `False`):
            Allows the user to decide whether or not to geographically
            project the data. If `True`, the other options in this tab are
            enabled.
        2. ``alpha`` (default 0.7):
            To adjust the opacity of the quadmesh on the map or projection.
        3. ``basemap`` (default `None`):
            To select basemap for the data from the available `tile_sources`_.
        4. ``crs`` (default `PlateCarree`):
            Here `crs` refers to `cartopy coordinate reference system`_, which
            declares the coordinate system of the data. It specifies the input
            projection, i.e. it declares how to interpret the incoming data
            values. It allows the user to select the `crs` in which data is
            present.
        5. ``crs params``:
            The default input parameters for each `crs` are auto filled upon
            selection. It allows the user to customize certain aspects of the
            projections by providing extra parameters such as
            ``central_longitude`` and ``central_latitude``. For more details
            refer to `cartopy coordinate reference system`_.
        6. ``projection`` (default `None`):
            To select what coordinate system to display the data in.
            We can also state it as output projection, i.e. how you want to
            map the data points onto the screen for display.
        7. ``projection params``:
            The default input parameters for each `projection` are auto filled
            upon selection. It allows the user to customize certain aspects of
            the projections by providing extra parameters such as
            ``central_longitude`` and ``central_latitude``.
        8. ``project`` (default `False`):
            Whether to project the data before plotting (adds initial overhead
            but avoids projecting data when plot is dynamically updated).
        9. ``global_extent`` (default `False`):
            Whether to expand the plot extent to span the whole globe.
        10. ``features`` (default=all except `None`):
            To select a set of basic geographic features to overlay behind the
            data in the plot.

        .. note::
            1. The widgets in this tab are enabled only if both ``x`` and ``y`` are data coordinates.
            2. `basemap` and `projection` are mutually exclusive options i.e it is not possible to have them both enabled simultaneously.


        .. _`cartopy projection`: https://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html
        .. _`Geoviews`: http://geoviews.org/
        .. _`tile_sources`: http://geoviews.org/gallery/bokeh/tile_sources.html
        .. _`cartopy coordinate reference system`: https://scitools.org.uk/cartopy/docs/v0.15/crs/index.html#coordinate-reference-systems-in-cartopy

    """

    def __init__(self):
        """Initializes the Projection pane."""
        super().__init__()
        self.is_geo = pn.widgets.Checkbox(name='is_geo', value=False,
                                          disabled=True)
        self.alpha = pn.widgets.FloatSlider(name='alpha', start=0, end=1,
                                            step=0.01, value=0.7, width=180)

        basemap_opts = {None: None}
        basemap_opts.update(gvts.tile_sources)
        self.basemap = pn.widgets.Select(name='basemap',
                                         options=basemap_opts,
                                         value=None)
        self.projection = pn.widgets.Select(name='projection',
                                            options=[None] + sorted(projections_list),
                                            value=None)
        self.crs = pn.widgets.Select(name='crs',
                                     options=sorted(projections_list),
                                     value='PlateCarree')
        self.project = pn.widgets.Checkbox(name='project', value=False)
        self.global_extent = pn.widgets.Checkbox(name='global_extent',
                                                 value=False)
        self.crs_params = pn.widgets.TextInput(name='crs params',
                                               value="{}", width=400)
        self.proj_params = pn.widgets.TextInput(name='projection params',
                                                value="{}", width=400)

        self.feature_ops = ['None', 'borders', 'coastline', 'grid', 'land',
                            'lakes', 'ocean', 'rivers']
        self.features = pn.widgets.MultiSelect(name='features',
                                               options=self.feature_ops,
                                               value=self.feature_ops[1:])

        self._register(self.is_geo, 'geo_changed')
        self._register(self.is_geo, 'geo_disabled', 'disabled')
        self._register(self.crs, 'add_crs_params')
        self._register(self.projection, 'add_proj_params')
        self._register(self.basemap, 'show_basemap')

        self.connect('geo_changed', self.setup)
        self.connect('geo_disabled', self.setup)
        self.connect('add_crs_params', self.add_crs_params)
        self.connect('add_proj_params', self.add_proj_params)
        self.connect('show_basemap', self.show_basemap)

        self.panel = pn.Column(pn.Row(self.is_geo),
                               pn.Row(self.alpha, self.basemap),
                               pn.Row(self.crs, self.crs_params, name='crs'),
                               pn.Row(self.projection, self.proj_params,
                                      name='proj'),
                               pn.Row(self.project,
                                      self.global_extent),
                               pn.Row(self.features),
                               name='Projection')
        self.setup()
        self.add_crs_params(self.crs.value)
        self.add_proj_params(self.projection.value)

    def setup(self, *args):
        # disable the widgets if is_geo is disabled or if is_geo is False
        if self.is_geo.disabled:
            disabled = True
        else:
            disabled = False if self.is_geo.value else True
        for row in self.panel[1:]:
            for widget in row:
                widget.disabled = disabled
        self.proj_params.disabled = disabled

    def setup_initial_values(self, init_params={}):
        """
        To select initial values for the widgets in this pane.
        """
        is_geo = init_params.get('is_geo')
        if is_geo:  # since we need to enable is_geo if True
            self.is_geo.disabled = False
        for row in self.panel:
            for widget in row:
                w_name = widget.name
                if w_name in init_params:
                    if w_name == 'basemap':
                        tile = init_params[w_name]
                        widget.value = getattr(gvts, tile) if tile is not None else None
                    else:
                        widget.value = init_params[w_name]

    def show_basemap(self, *args):
        value = False if self.basemap.value is None else True
        self.projection.disabled = value
        self.proj_params.disabled = value
        self.features.value = [self.feature_ops[0]] if value else self.feature_ops[1:]

    def add_crs_params(self, *args):
        if args[0] != None:
            projs = proj_params(args[0])
            self.crs_params.value = projs

    def add_proj_params(self, *args):
        if args[0] != None:
            projs = proj_params(args[0])
            self.proj_params.value = projs

    def disable_geo(self, value):
        self.is_geo.disabled = value

    @property
    def kwargs(self):
        out = {widget.name: widget.value
               for row in self.panel for widget in row}
        return out



import panel as pn
import pytest
from panel.widgets import DiscretePlayer, DiscreteSlider, Select
from ..utils import convert_widget, player_with_name_and_value


@pytest.mark.parametrize("source_widget,target_widget",
                         [(Select, DiscretePlayer),
                          (DiscreteSlider, Select),
                          (DiscretePlayer, Select),
                          (DiscreteSlider, DiscretePlayer)])
def test_convert_widget(source_widget, target_widget):
    name = 'Alphabets'
    options = ['A', 'B', 'C', 'D']
    value = options[0]
    source = source_widget(name=name,
                           options=options,
                           value=value)

    target = convert_widget(source, target_widget())
    assert isinstance(target, target_widget)
    assert target.name is source.name
    assert target.options is source.options

    # change target's value and check if source's also change
    target.value = options[1]
    assert target.value is source.value


def test_player_with_name_and_value():
    options = ['a', 'b', 'c']
    name = "Player1"
    player = pn.widgets.DiscretePlayer(name=name,
                                       value=options[0],
                                       options=options)
    out = player_with_name_and_value(player)
    out_markdown_name = out[0][0]
    assert out_markdown_name.object == player.name
    # Upon changing the value of player, markdown also changes to
    # represent the same
    player.value = 'c'
    out_markdown_value = out[0][1]
    assert out_markdown_value.object == player.value



#!/usr/bin/env python

from setuptools import setup, find_packages
import versioneer


requires = open('requirements.txt').read().strip().split('\n')

setup(
    name='xrviz',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Interactive visualisations for xarrays',
    url='',
    maintainer='Martin Durant',
    maintainer_email='mdurant@anaconda.com',
    license='BSD',
    py_modules=['xrviz'],
    packages=find_packages(),
    package_data={'': ['*.csv', '*.yml', '*.html']},
    include_package_data=True,
    install_requires=requires,
    long_description=open('README.md').read(),
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'xrviz = xrviz.cli:main'
        ]},
)



import logging

logger = logging.getLogger('xrviz')

try:
    from cartopy import crs as ccrs
    import geoviews as gv
    import geoviews.feature as gf
    has_cartopy = True
except ImportError:
    logger.debug("Install Cartopy, Geoviews to view Projection Panel.")
    has_cartopy = False
    ccrs, gv, gf = None, None, None


try:
    import metpy.calc as mpcalc
except ImportError:
    logger.debug("""
    Metpy is not installed, so XrViz cannot automatically guess the best
    dimensions for output x and y axes. To install metpy, you can execute
    the following:
        `pip install metpy`
            or
        `conda install -c conda-forge metpy pint pooch --no-deps`
    """)
    mpcalc = None

try:
    import crick.tdigest
    has_crick_tdigest = True
except ImportError:
    logger.debug("Install crick to use tdigest method for finding cmap limits.")
    has_crick_tdigest = False



import pytest
from xrviz.coord_setter import CoordSetter
from . import data

@pytest.fixture()
def coord_setter(data):
    return CoordSetter(data)


def test_coord_setter_initial(coord_setter, data):
    assert coord_setter.panel.name == 'Set Coords'
    assert coord_setter.coord_selector.options == list(data.variables)
    assert coord_setter.coord_selector.value == list(data.coords)



import sys

def main():
    usage = "Usage: `xrviz show datafile/url`"
    if len(sys.argv) ==3 and 'show' == sys.argv[1]:
        import xarray as xr
        from .dashboard import Dashboard

        try:
            data = xr.open_dataset(sys.argv[2])
        except:
            print("Unable to open the datafile/url.", usage)
            sys.exit()
        dash = Dashboard(data)
        dash.show()
    else:
        print(usage)



import panel as pn
import xarray as xr
from .sigslot import SigSlot
from .display import Display
from .describe import Describe
from .fields import Fields
from .style import Style
from .coord_setter import CoordSetter
from .compatibility import has_cartopy


class Control(SigSlot):
    """
    The user input part of the interface

    Parameters
    ----------
    data: xarray.DataSet
        The data to be visualised. Here, we are mostly concerned with
        displaying the variables, their attributes, and assigning coordinates
        to roles upon plotting.

    Attributes
    ----------
    1. panel:
            A ``panel.Tabs`` instance containing the user input panes
    2. displayer:
            A ``Display`` instance, displays a list of data variables for selection.
    3. describer:
            A ``Describe`` instance, describes the properties of the variable
            selected in the ``displayer``.
    4. coord_setter:
            A ``CoordSetter`` instance for choosing which variables are
            considered coordinates.
    5. fields:
            A ``Fields`` instance to select the axes to plot with.
    6. style:
            A ``Style`` instance to customise styling of the graphs.
    7. projection:
            A ``Projection`` instance to customise the projection of
            geographical data.
    8. kwargs:
            A dictionary gathered from the widgets of the input Panes,
            of a form which can be passed to the plotting function as kwargs.
    """

    def __init__(self, data):
        super().__init__()
        self.data = data
        self.displayer = Display(self.data)
        self.describer = Describe(self.data)
        self.fields = Fields(self.data)
        self.style = Style()
        self.coord_setter = CoordSetter(self.data)
        self.tabs = pn.Tabs(
            pn.Row(self.displayer.panel, self.describer.panel,
                   name='Variables', width_policy='max'),
            self.coord_setter.panel,
            self.fields.panel,
            self.style.panel,
            background='#f5f5f5',
            width_policy='max',
            margin=20
        )

        if has_cartopy:
            from .projection import Projection
            self.projection = Projection()
            self.tabs.append(self.projection.panel)
            self.fields.connect('x', self.check_is_projectable)
            self.fields.connect('y', self.check_is_projectable)

        self.displayer.connect("variable_selected", self.describer.setup)
        self.displayer.connect("variable_selected", self.fields.setup)
        self.displayer.connect("variable_selected", self.style.setup)

        self.panel = pn.WidgetBox(self.tabs, width_policy='max')

    def setup_initial_values(self, initial_params={}):
        self.displayer.setup_initial_values(initial_params)
        self.coord_setter.setup_initial_values(initial_params)
        self.fields.setup_initial_values(initial_params)
        self.style.setup_initial_values(initial_params)
        if has_cartopy:
            self.projection.setup_initial_values(initial_params)

    def set_coords(self, data):
        """
        Called after coordinates have been set, to update the other input panes.
        """
        try:  # Upon setting coords before selecting a variable
            var = self.kwargs['Variables']
        except:
            var = None
        self.data = data
        self.coord_setter.set_coords(self.data)
        self.displayer.set_coords(self.data)
        self.describer.set_coords(self.data, var)
        self.fields.set_coords(self.data, var)

    def check_is_projectable(self, *args):
        """
        Check if the selected variable can be  projected geographically.

        This is possible only when both `x` and `y` are present in selected
        variable's coordinates.
        """
        value = not self.fields.kwargs['are_var_coords']
        self.projection.disable_geo(value)

    @property
    def kwargs(self):
        out = self.displayer.kwargs
        out.update(self.fields.kwargs)
        out.update(self.style.kwargs)
        if has_cartopy:
            out.update(self.projection.kwargs)
        return out
